package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.*;
import dev.pfaff.altargui.function.IndexedFunction;
import dev.pfaff.altargui.handlers.MouseScrollHandler;
import dev.pfaff.altargui.inspector.DebugLayer;
import org.jetbrains.annotations.NotNull;

import java.util.List;
import java.util.function.BiConsumer;

public record FiniteScrollListWidget(int count, KeyFunction keyFunction, Builder builder) implements Widget {
	public static final DebugLayer DEBUG_LAYER_PRE_UNSCISSORED = new DebugLayer(FiniteScrollListWidget.class, "pre_unscissored");
	public static final DebugLayer DEBUG_LAYER_PRE_SCISSORED = new DebugLayer(FiniteScrollListWidget.class, "pre_scissored");
	public static final DebugLayer DEBUG_LAYER_POST_UNSCISSORED = new DebugLayer(FiniteScrollListWidget.class, "post_unscissored");

	public static <T> FiniteScrollListWidget create(List<T> items,
													IndexedFunction<T, Key.CallSite> keyFunction,
													IndexedFunction<T, Widget> builder) {
		return new FiniteScrollListWidget(items.size(),
				i -> keyFunction.apply(i, items.get(i)),
				i -> builder.apply(i, items.get(i)));
	}

	@Override
	public @NotNull WidgetInstance instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	@FunctionalInterface
	public interface KeyFunction {
		Key.CallSite apply(int index);
	}

	@FunctionalInterface
	public interface Builder {
		Widget build(int index);
	}

	private static final class State extends WidgetInstance<FiniteScrollListWidget> implements MouseScrollHandler {
		private int count;
		private KeyFunction keyFunction;
		private Builder builder;

		/**
		 * The index of the first item displayed.
		 */
		private int itemI;

		/**
		 * The offset of the first item displayed.
		 */
		private double itemOffset;

		private final WidgetInstanceList children = new WidgetInstanceList(2);

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		public void configure(StateContext stateContext, FiniteScrollListWidget widget, boolean init) {
			count = widget.count;
			keyFunction = widget.keyFunction;
			builder = widget.builder;
		}

		// TODO: cache built widgets
		@Override
		protected void computeLayout(LayoutContext context) {
			var children = this.children;
			children.clear();

			var builder = this.builder;
			var keyFunction = this.keyFunction;
			int count = this.count;
			final float x = this.allocatedX();
			final float minY = this.allocatedY();
			double y = minY;
			final var allocWidth = this.allocatedWidth();
			final var allocHeight = this.allocatedHeight();
			float width = 0;
			var maxY = minY + allocHeight;
			y += this.itemOffset;
			eachElement: for (int i = this.itemI; i < count && y < maxY; ) {
				var widget = builder.build(i);
				{

					var availHeight = (float)(maxY - y);
					int i_ = i;
					var element = context.getState(this, keyFunction.apply(i), widget);
					var y_ = y;
					var layouter = new BiConsumer<WidgetInstance<?>, LayoutContext>() {
						private double ySlot = y_;

						@Override
						public void accept(WidgetInstance<?> e, LayoutContext ctx) {
							e.setAllocatedRect(x, (float) this.ySlot, allocWidth, availHeight, false, true);
							e.layout(ctx);

							float height = e.computedHeight();
							if (i_ == State.this.itemI && i_ == count - 1) {
								if (this.ySlot < minY && this.ySlot + height < maxY) {
									this.ySlot = minY + (State.this.itemOffset = Math.max(0, minY - (maxY - height)));

									e.setAllocatedRect(x, (float) this.ySlot, allocWidth, availHeight, false, true);
									e.layout(ctx);
								}
							} else if (this.ySlot + height <= minY) {
								if (i_ + 1 < count) {
									State.this.itemI = i_ + 1;
									State.this.itemOffset += height;
								}
							}
						}
					};
					context.layout(this, element, layouter);
					float height = element.computedHeight();
					if (i == this.itemI && y > minY) {
						if (i > 0) {
							this.itemI = --i;
							this.itemOffset -= height;
							y -= height;
						} else {
							this.itemOffset = 0;
							y = minY;
						}
						continue eachElement;
					}
					children.add(element);
					y = layouter.ySlot;
					width = Math.max(width, element.computedWidth());
					y += height;
				}

				i++;
			}

			setComputedSize(width, allocHeight);
		}

		@Override
		protected void computeMount(MountContext context) {
			context.pushScissor(this);
			context.mountList(children);
			context.popScissor();
		}

		@Override
		public void uSpecMount(MountContext context) {
			super.uSpecMount(context);
		}

		@Override
		public boolean mouseScrolled(double mouseX, double mouseY, double delta) {
			if (isInside(mouseX, mouseY)) {
				itemOffset += MouseScrollHandler.applyStandardDeltaTransform(delta);
				markNeedsLayout();
				return true;
			}
			return false;
		}
	}
}
