package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.*;
import dev.pfaff.altargui.inspector.DebugLayer;
import dev.pfaff.altargui.selector.BuiltinSelector;
import dev.pfaff.altargui.selector.Children;
import dev.pfaff.altargui.selector.NthElement;
import dev.pfaff.altargui.util.ListMultiChildBuilder;
import org.jetbrains.annotations.NotNull;

import java.util.List;
import java.util.function.BiConsumer;
import java.util.function.Consumer;

/**
 * Builder is only <em>guaranteed</em> to be called when it changes.
 */
public record MultiChildWidget(@NotNull Mode mode, Builder builder) implements Widget {
	public static final DebugLayer DEBUG_LAYER = new DebugLayer(MultiChildWidget.class);

	private static final MultiChildWidget EMPTY_STACK = new MultiChildWidget(Mode.Stack, ignored -> {});
	private static final MultiChildWidget EMPTY_COL = new MultiChildWidget(Mode.Column, ignored -> {});
	private static final MultiChildWidget EMPTY_ROW = new MultiChildWidget(Mode.Row, ignored -> {});
	private static final MultiChildWidget EMPTY_DIAGONAL = new MultiChildWidget(Mode.Diagonal, ignored -> {});

	public enum Mode {
		Stack,
		Row,
		Column,
		Diagonal,
	}

	public static MultiChildWidget stack(Builder builder) {
		return new MultiChildWidget(Mode.Stack, builder);
	}

	public static MultiChildWidget row(Builder builder) {
		return new MultiChildWidget(Mode.Row, builder);
	}

	public static MultiChildWidget col(Builder builder) {
		return new MultiChildWidget(Mode.Column, builder);
	}

	public static MultiChildWidget diagonal(Builder builder) {
		return new MultiChildWidget(Mode.Diagonal, builder);
	}

	public static MultiChildWidget stack(List<? extends Widget> children) {
		return stack(ListMultiChildBuilder.newUntrusted(children));
	}

	public static MultiChildWidget row(List<? extends Widget> children) {
		return row(ListMultiChildBuilder.newUntrusted(children));
	}

	public static MultiChildWidget col(List<? extends Widget> children) {
		return col(ListMultiChildBuilder.newUntrusted(children));
	}

	public static MultiChildWidget diagonal(List<? extends Widget> children) {
		return diagonal(ListMultiChildBuilder.newUntrusted(children));
	}

	public static MultiChildWidget emptyStack() {
		return EMPTY_STACK;
	}

	public static MultiChildWidget emptyCol() {
		return EMPTY_COL;
	}

	public static MultiChildWidget emptyRow() {
		return EMPTY_ROW;
	}

	public static MultiChildWidget emptyDiagonal() {
		return EMPTY_DIAGONAL;
	}

	@Override
	public @NotNull WidgetInstance instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	@FunctionalInterface
	public interface Builder extends Consumer<ManagedLayoutContext> {
		@Override
		void accept(ManagedLayoutContext context);

		@NotNull
		@Override
		default Builder andThen(@NotNull Consumer<? super ManagedLayoutContext> after) {
			return context -> {
				this.accept(context);
				after.accept(context);
			};
		}
	}

	private static final class State extends WidgetInstance<MultiChildWidget> {
		private @NotNull Mode mode;
		private Builder builder;
		private final WidgetInstanceList children = new WidgetInstanceList(2);

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		public void configure(StateContext stateContext, MultiChildWidget widget, boolean init) {
			mode = widget.mode;
			if (!widget.builder.equals(builder)) {
				builder = widget.builder;
				children.clear();
				builder.accept((key, widget1) -> {
					children.add(stateContext.getState(this, key, widget1));
				});
				markNeedsLayout();
			}
		}

		@Override
		protected void computeLayout(LayoutContext context) {
			var children = this.children;
			var mode = this.mode;
			final float minX = this.allocatedX();
			final float minY = this.allocatedY();
			final float allocWidth = this.allocatedWidth();
			final var allocHeight = this.allocatedHeight();
			switch (mode) {
				case Stack -> {
					var layoutFunction = new StackLayoutFunction(minX, minY, allocWidth, allocHeight, false);
					children.forEach(child -> context.layout(this, child, layoutFunction));
					if (layoutFunction.anyTight) {
						children.forEach(child -> {
							if (child instanceof SizedWidget.TightElement tightChild) {
								float width_ = tightChild.isX() ? layoutFunction.compWidth : allocWidth;
								float height_ = tightChild.isY() ? layoutFunction.compHeight : allocHeight;
								var tightLayoutFunction = new StackLayoutFunction(minX, minY, width_, height_, true);
								tightLayoutFunction.accept((WidgetInstance) tightChild, context);
								layoutFunction.compWidth = Math.max(layoutFunction.compWidth,
																	tightLayoutFunction.compWidth);
								layoutFunction.compHeight = Math.max(layoutFunction.compHeight,
																	 tightLayoutFunction.compHeight);
							}
						});
					}
					setComputedSize(layoutFunction.compWidth, layoutFunction.compHeight);
				}
				case Row -> {
					var layouter = new BiConsumer<WidgetInstance<?>, LayoutContext>() {
						private float offX = 0f;
						private float maxHeight = 0f;

						@Override
						public void accept(WidgetInstance e, LayoutContext ctx) {
							e.setAllocatedRect(minX + this.offX, minY, allocWidth - this.offX, allocHeight);
							e.layout(ctx);
							this.offX += e.computedWidth();
							this.maxHeight = Math.max(this.maxHeight, e.computedHeight());
						}
					};
					children.forEach(child -> context.layout(this, child, layouter));
					setComputedSize(layouter.offX, layouter.maxHeight);
				}
				case Column -> {
					var layouter = new BiConsumer<WidgetInstance<?>, LayoutContext>() {
						private float maxWidth = 0f;
						private float offY = 0f;

						@Override
						public void accept(WidgetInstance e, LayoutContext ctx) {
							e.setAllocatedRect(minX, minY + this.offY, allocWidth, allocHeight - this.offY);
							e.layout(ctx);
							this.maxWidth = Math.max(this.maxWidth, e.computedWidth());
							this.offY += e.computedHeight();
						}
					};
					children.forEach(child -> context.layout(this, child, layouter));
					setComputedSize(layouter.maxWidth, layouter.offY);
				}
				case Diagonal -> {
					var layouter = new BiConsumer<WidgetInstance<?>, LayoutContext>() {
						private float offX = 0f;
						private float offY = 0f;

						@Override
						public void accept(WidgetInstance<?> e, LayoutContext ctx) {
							e.setAllocatedRect(minX + this.offX,
									minY + this.offY,
									allocWidth - this.offX,
									allocHeight - this.offY);
							e.layout(ctx);
							this.offX += e.computedWidth();
							this.offY += e.computedHeight();
						}
					};
					children.forEach(child -> context.layout(this, child, layouter));
					setComputedSize(layouter.offX, layouter.offY);
				}
			}
		}

		@Override
		protected void computeMount(MountContext context) {
			context.mountList(children);
		}

		@Override
		public void uSpecMount(MountContext context) {
			super.uSpecMount(context);
		}

		@Override
		public void applyBuiltinSelector(BuiltinSelector selector, Consumer<WidgetInstance> consumer) {
			switch (selector) {
				case NthElement nthElement -> {
					int n = nthElement.n();
					if (n < 0) n = this.children.size() - n;
					if (n >= 0 && n < this.children.size()) {
						BuiltinSelector.applyNext(this.children.get(n),
												  nthElement,
												  consumer);
					}
				}
				case Children children -> {
					this.children.forEach(child -> BuiltinSelector.applyNext(child, children, consumer));
				}
			}
		}
	}

	private static final class StackLayoutFunction implements BiConsumer<WidgetInstance<?>, LayoutContext> {
		private final float x;
		private final float y;
		private final float width;
		private final float height;
		private final boolean tightPass;
		float compWidth = 0;
		float compHeight = 0;
		boolean anyTight = false;

		public StackLayoutFunction(float x, float y, float width, float height, boolean tightPass) {
			this.x = x;
			this.y = y;
			this.width = width;
			this.height = height;
			this.tightPass = tightPass;
		}

		@Override
		public void accept(WidgetInstance e, LayoutContext ctx) {
			float x = this.x;
			float y = this.y;
			float width = this.width;
			float height = this.height;
			if (tightPass) {
				var tightE = (SizedWidget.TightElement) e;
			} else {
				if (e instanceof SizedWidget.TightElement) {
					anyTight = true;
					return;
				}
			}
			e.setAllocatedRect(x, y, width, height);
			e.layout(ctx);

			compWidth = Math.max(compWidth, e.computedWidth());
			compHeight = Math.max(compHeight, e.computedHeight());
		}
	}
}
