package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.*;
import dev.pfaff.altargui.types.Alignment;
import org.jetbrains.annotations.NotNull;

public sealed interface PositionedWidget extends SingleChildWidget {
	public record Absolute(float x, float y, Widget child) implements PositionedWidget {
		@Override
		public @NotNull WidgetInstance instantiate(WidgetInstantiationArgs args) {
			return new State(args);
		}

		private static final class State extends CustomLayoutState<Absolute> {
			private float x;
			private float y;

			private State(WidgetInstantiationArgs args) {
				super(args);
			}

			@Override
			public void configure(StateContext stateContext, Absolute widget, boolean init) {
				super.configure(stateContext, widget, init);
				if (x != widget.x || y != widget.y) {
					x = widget.x;
					y = widget.y;
					markNeedsLayout();
				}
			}

			@Override
			protected void computeLayout(LayoutContext context) {
				context.layout(this, getChildElement(), (element, ctx) -> {
//					var allocWidth = mcWindow.getScaledWidth();
//					var allocHeight = mcWindow.getScaledHeight();
					var allocWidth = this.allocatedWidth();
					var allocHeight = this.allocatedHeight();
					element.setAllocatedRect(this.x, this.y, allocWidth - this.x, allocHeight - this.y, true, true);
					element.layout(ctx);
					setComputedSize(0, 0);
				});
			}
		}
	}

	public record Relative(float x, float y, Widget child) implements PositionedWidget {
		@Override
		public @NotNull WidgetInstance instantiate(WidgetInstantiationArgs args) {
			return new State(args);
		}

		private static final class State extends CustomLayoutState<Relative> {
			private float x;
			private float y;

			private State(WidgetInstantiationArgs args) {
				super(args);
			}

			@Override
			public void configure(StateContext stateContext, Relative widget, boolean init) {
				super.configure(stateContext, widget, init);
				if (x != widget.x || y != widget.y) {
					x = widget.x;
					y = widget.y;
					markNeedsLayout();
				}
			}

			@Override
			protected void computeLayout(LayoutContext context) {
				context.layout(this, getChildElement(), (element, ctx) -> {
					boolean canOverflowX = canOverflowX();
					boolean canOverflowY = canOverflowY();
					element.setAllocatedRect(allocatedX() + x,
							allocatedY() + y,
							allocatedWidth() - x,
							allocatedHeight() - y,
							canOverflowX,
							canOverflowY);
					element.layout(ctx);
					setComputedSize(Math.max(0f, x + element.computedWidth()), Math.max(0f, y + element.computedHeight()));
				});
			}
		}
	}

	public record Aligned(Widget child, Alignment horizontal, Alignment vertical) implements PositionedWidget {
		@Override
		public @NotNull WidgetInstance instantiate(WidgetInstantiationArgs args) {
			return new State(args);
		}

		private static final class State extends CustomLayoutState<Aligned> {
			private @NotNull Alignment horizontal;
			private @NotNull Alignment vertical;

			private State(WidgetInstantiationArgs args) {
				super(args);
			}

			@Override
			public void configure(StateContext stateContext, Aligned widget, boolean init) {
				super.configure(stateContext, widget, init);
				if (horizontal != widget.horizontal || vertical != widget.vertical) {
					horizontal = widget.horizontal;
					vertical = widget.vertical;
					markNeedsLayout();
				}
			}

			@Override
			protected void computeLayout(LayoutContext context) {
				var horizontal = this.horizontal;
				var vertical = this.vertical;
				context.layout(this, getChildElement(), (e, ctx) -> {
					e.copyAllocatedRectFrom(this);
					e.layout(ctx);
					var allocX = allocatedX();
					var allocY = allocatedY();
					var allocWidth = allocatedWidth();
					var allocHeight = allocatedHeight();
					// TODO: do this without 2 passes
					e.setAllocatedRect(allocX + horizontal.apply(e.computedWidth(), allocWidth),
							allocY + vertical.apply(e.computedHeight(), allocHeight),
							allocWidth,
							allocHeight,
							canOverflowX(),
							canOverflowY());
					e.layout(ctx);
					setComputedSize(switch (horizontal) {
						case Start -> e.allocatedX() - allocX + e.computedWidth();
						case Center, End -> allocWidth;
					}, switch (vertical) {
						case Start -> e.allocatedY() - allocY + e.computedHeight();
						case Center, End -> allocHeight;
					});
				});
			}
		}
	}
}
