package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.*;
import dev.pfaff.altargui.element.Painter;
import dev.pfaff.altargui.render.objects.IRectRenderObject;
import dev.pfaff.altargui.render.objects.RectRenderObject;
import dev.pfaff.altargui.render.objects.TexturedRectRenderObject;
import dev.pfaff.altargui.types.Color;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.Objects;

/**
 * @param colorRgba the RGBA colorRgba of the widget
 * @param tex the texture of the widget
 */
public record RectWidget(int colorRgba, @Nullable LoadableTex2d tex) implements Widget {
	public static final RectWidget TRANSPARENT = new RectWidget(0, null);
	public static final RectWidget GENERIC_HUD_BACKGROUND = colored(Color.fromArgb(0x90_505050));
	public static final RectWidget GENERIC_HEAVY_HUD_BACKGROUND = colored(Color.fromArgb(0xcc_505050));

	/**
	 * @deprecated treat this as private.
	 */
	@Deprecated
	public RectWidget {
	}

	/**
	 * @param color ARGB
	 */
	public static RectWidget colored(Color color) {
		if (color.alpha() == 0) return TRANSPARENT;
		return new RectWidget(color.rgba(), null);
	}

	public static RectWidget textured(@NotNull LoadableTex2d tex) {
		return new RectWidget(Color.WHITE.rgba(), Objects.requireNonNull(tex));
	}

	@Override
	public @NotNull WidgetInstance<RectWidget> instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	public Color color() {
		return Color.fromRgba(colorRgba);
	}

	public RectWidget withColor(Color color) {
		return new RectWidget(color.rgba(), tex);
	}

	public RectWidget withTexture(LoadableTex2d tex) {
		return new RectWidget(colorRgba, tex);
	}

	private static final class State extends WidgetInstance<RectWidget> implements Painter {
		private IRectRenderObject rect = null;
		private int color;
		private @Nullable LoadableTex2d tex;
		private boolean loaded = true;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		public void configure(StateContext stateContext, RectWidget widget, boolean init) {
			if (!(widget.tex != null ? rect instanceof TexturedRectRenderObject : rect instanceof RectRenderObject)) {
				rect = widget.tex != null ? new TexturedRectRenderObject() : new RectRenderObject();
			}
			if (widget.colorRgba != this.color) {
				this.rect.updateColor(Color.fromRgba(this.color = widget.colorRgba));
			}
			if (!Objects.equals(widget.tex, this.tex)) {
				this.tex = widget.tex;
				if (this.tex != null) {
					this.loaded = false;
				} else {
					this.loaded = true;
					if (rect instanceof TexturedRectRenderObject textured) {
						// the white pixel texture should be locked on program 0 at 0, 0, 1, 1
//						textured.updateTexLayer(0);
//						textured.updateUV(0f, 0f, 0f, 0f);
						// FIXME
					}
				}
			}
		}

		@Override
		protected void computeLayout(LayoutContext context) {
			computeFillLayout(this, context);
			float x1 = this.allocatedX();
			float y1 = this.allocatedY();
			float width = this.computedWidth();
			float height = this.computedHeight();
			this.rect.updatePositionSize(x1, y1, width, height);
		}

		@Override
		protected void computeMount(MountContext context) {
		}

		@Override
		public void uSpecMount(MountContext context) {
			super.uSpecMount(context);
		}

		@Override
		public void paint(long time, Canvas canvas) {
			if (Color.fromRgba(this.color).alpha() == 0) return;

			var rect = this.rect;

			if (!this.loaded) {
				this.loaded = true;
				var tex = Objects.requireNonNull(this.tex);
				var coords = tex.coords();
				canvas.texAlloc().alloc(tex.tex()).thenAccept(handle -> {
					if (rect instanceof TexturedRectRenderObject textured) {
						textured.updateTex(handle, coords);
					}
				});
			}

			rect.uSpecUploadCanvas((CanvasImpl) canvas);
			//canvas.upload(rect);
		}
	}
}
