package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.StateContext;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.WidgetInstance;
import dev.pfaff.altargui.WidgetInstantiationArgs;
import dev.pfaff.altargui.element.StatefulWidgetInstance;
import dev.pfaff.altargui.util.listenable.ListenerDisposer;
import dev.pfaff.altargui.util.listenable.ValueListenable;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.function.Function;

/**
 * A widget that replaces its child using the provided {@code builder} whenever notified by listenable.
 */
public record ValueListenableBuilderWidget(ValueListenable<@NotNull Widget> listenable) implements Widget {
	public static <T> ValueListenableBuilderWidget builder(ValueListenable<T> listenable, Function<T, @NotNull Widget> builder) {
		return new ValueListenableBuilderWidget(listenable.map(builder));
	}

	@Override
	public @NotNull WidgetInstance<ValueListenableBuilderWidget> instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	private static final class State extends StatefulWidgetInstance<ValueListenableBuilderWidget> {
		private ValueListenable<@NotNull Widget> listenable;
		private Widget child;

		private @Nullable ListenerDisposer listener;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		protected @NotNull Widget build() {
			return child;
		}

		@Override
		public void configure(StateContext stateContext, ValueListenableBuilderWidget widget, boolean init) {
			if (listenable != widget.listenable) {
				if (listener != null) listener.close();
				var listenable = this.listenable = widget.listenable;
				listener = listenable.addListener(child -> {
					//noinspection ConstantValue
					if (child == null) {
						// using this.listenable here to avoid another capture.
						throw new NullPointerException(STR."\{ValueListenableBuilderWidget.class.getName()} listenable emitted a null value: \{this.listenable}");
					}
					this.child = child;
					this.markNeedsRebuild();
				});
			}
		}

		@Override
		public void dispose() {
			if (listener != null) {
				listener.close();
				listener = null;
			}
			super.dispose();
		}
	}
}
